import json
from logging import fatal

from mm_std import BaseConfig, Err, print_plain
from pydantic import StrictStr

from mm_eth import abi, rpc
from mm_eth.cli.cli_utils import BaseConfigParams


class Config(BaseConfig):
    contract_address: StrictStr
    function_signature: str
    function_args: StrictStr = "[]"
    outputs_types: str | None = None
    node: str


class CallContractCmdParams(BaseConfigParams):
    pass


def run(cli_params: CallContractCmdParams) -> None:
    config = Config.read_toml_config_or_exit(cli_params.config_path)
    if cli_params.print_config:
        config.print_and_exit()

    input_data = abi.encode_function_input_by_signature(
        config.function_signature,
        json.loads(config.function_args.replace("'", '"')),
    )
    res = rpc.eth_call(config.node, config.contract_address, input_data)
    if isinstance(res, Err):
        return fatal(f"error: {res.err}")

    result = res.ok
    if config.outputs_types is not None:
        decode_res = abi.decode_data(_get_types(config.outputs_types), result)
        result = decode_res[0] if len(decode_res) == 1 else str(decode_res)
    print_plain(result)


def _get_types(data: str) -> list[str]:
    return [t.strip() for t in data.split(",") if t.strip()]



import mm_crypto_utils
from loguru import logger
from mm_crypto_utils import VarInt, get_log_prefix
from mm_std import Err, fatal

from mm_eth import erc20, rpc
from mm_eth.utils import from_wei_str

from .calcs import calc_eth_expression


def get_nonce(nodes: list[str] | str, address: str, log_prefix: str | None = None) -> int | None:
    res = rpc.eth_get_transaction_count(nodes, address, attempts=5)
    prefix = log_prefix or address
    logger.debug(f"{prefix}: nonce={res.ok_or_err()}")
    if isinstance(res, Err):
        logger.info(f"{prefix}: nonce error, {res.err}")
        return None
    return res.ok


def check_nodes_for_chain_id(nodes: list[str], chain_id: int) -> None:
    for node in nodes:
        res = rpc.eth_chain_id(node, timeout=7)
        if isinstance(res, Err):
            fatal(f"can't get chain_id for {node}, error={res.err}")
        if res.ok != chain_id:
            fatal(f"node {node} has a wrong chain_id: {res.ok}")


def get_base_fee(nodes: list[str], log_prefix: str | None = None) -> int | None:
    res = rpc.get_base_fee_per_gas(nodes)
    prefix = get_log_prefix(log_prefix)
    logger.debug(f"{prefix}base_fee={res.ok_or_err()}")
    if isinstance(res, Err):
        logger.info(f"{prefix}base_fee error, {res.err}")
        return None
    return res.ok


def calc_max_fee(nodes: list[str], max_fee_expression: str, log_prefix: str | None = None) -> int | None:
    if "base_fee" in max_fee_expression.lower():
        base_fee = get_base_fee(nodes, log_prefix)
        if base_fee is None:
            return None
        return calc_eth_expression(max_fee_expression, VarInt("base_fee", base_fee))

    return calc_eth_expression(max_fee_expression)


def is_max_fee_limit_exceeded(max_fee: int, max_fee_limit_expression: str | None, log_prefix: str | None = None) -> bool:
    if max_fee_limit_expression is None:
        return False
    max_limit_value = calc_eth_expression(max_fee_limit_expression)
    if max_fee > max_limit_value:
        prefix = get_log_prefix(log_prefix)
        logger.info(
            "{}max_fee_limit is exceeded, max_fee={}, max_fee_limit={}",
            prefix,
            from_wei_str(max_fee, "gwei"),
            from_wei_str(max_limit_value, "gwei"),
        )
        return True
    return False


def calc_gas(
    *,
    nodes: list[str],
    gas_expression: str,
    from_address: str,
    to_address: str,
    value: int | None = None,
    data: str | None = None,
    log_prefix: str | None = None,
) -> int | None:
    var = None
    if "estimate" in gas_expression.lower():
        prefix = get_log_prefix(log_prefix)
        res = rpc.eth_estimate_gas(nodes, from_address, to_address, data=data, value=value, attempts=5)
        logger.debug(f"{prefix}gas_estimate={res.ok_or_err()}")
        if isinstance(res, Err):
            logger.info(f"{prefix}estimate_gas error, {res.err}")
            return None
        var = VarInt("estimate", res.ok)
    return calc_eth_expression(gas_expression, var)


def calc_eth_value_for_address(
    *,
    nodes: list[str],
    value_expression: str,
    address: str,
    gas: int,
    max_fee: int,
    log_prefix: str | None = None,
) -> int | None:
    var = None
    if "balance" in value_expression.lower():
        prefix = get_log_prefix(log_prefix)
        res = rpc.eth_get_balance(nodes, address, attempts=5)
        logger.debug(f"{prefix}balance={res.ok_or_err()}")
        if isinstance(res, Err):
            logger.info(f"{prefix}balance error, {res.err}")
            return None
        var = VarInt("balance", res.ok)

    value = calc_eth_expression(value_expression, var)
    if "balance" in value_expression.lower():
        value = value - gas * max_fee
    return value


def calc_erc20_value_for_address(
    *,
    nodes: list[str],
    value_expression: str,
    wallet_address: str,
    token_address: str,
    decimals: int,
    log_prefix: str | None = None,
) -> int | None:
    value_expression = value_expression.lower()
    var = None
    if "balance" in value_expression:
        prefix = get_log_prefix(log_prefix)
        res = erc20.get_balance(nodes, token_address, wallet_address, attempts=5)
        logger.debug(f"{prefix}balance={res.ok_or_err()}")
        if isinstance(res, Err):
            logger.info(f"{prefix}balance error, {res.err}")
            return None
        var = VarInt("balance", res.ok)
    return mm_crypto_utils.calc_int_expression(value_expression, var, suffix_decimals={"t": decimals})



from pathlib import Path

from mm_std import pretty_print_toml


def run(command: str) -> None:
    command = command.replace("-", "_")
    example_file = Path(Path(__file__).parent.absolute(), "../examples", f"{command}.toml")
    pretty_print_toml(example_file.read_text())



import sys
import time
from pathlib import Path
from typing import Annotated, Self

import mm_crypto_utils
from loguru import logger
from mm_crypto_utils import AddressToPrivate, Transfer
from mm_std import BaseConfig, Err, fatal, utc_now
from pydantic import AfterValidator, BeforeValidator, Field, model_validator
from rich.console import Console
from rich.live import Live
from rich.table import Table

from mm_eth import erc20, rpc
from mm_eth.cli import calcs, cli_utils, rpc_helpers
from mm_eth.cli.calcs import calc_eth_expression
from mm_eth.cli.cli_utils import BaseConfigParams
from mm_eth.cli.validators import Validators
from mm_eth.tx import sign_tx
from mm_eth.utils import from_wei_str


class Config(BaseConfig):
    nodes: Annotated[list[str], BeforeValidator(Validators.nodes())]
    chain_id: int
    transfers: Annotated[list[Transfer], BeforeValidator(Validators.eth_transfers())]
    private_keys: Annotated[AddressToPrivate, BeforeValidator(Validators.eth_private_keys())]
    token: Annotated[str | None, AfterValidator(Validators.eth_address())] = None  # if None, then eth transfer
    token_decimals: int = -1
    max_fee: Annotated[str, AfterValidator(Validators.valid_eth_expression("base_fee"))]
    priority_fee: Annotated[str, AfterValidator(Validators.valid_eth_expression())]
    max_fee_limit: Annotated[str | None, AfterValidator(Validators.valid_eth_expression())] = None
    default_value: Annotated[str | None, AfterValidator(Validators.valid_eth_or_token_expression("balance"))] = None
    value_min_limit: Annotated[str | None, AfterValidator(Validators.valid_eth_or_token_expression())] = None
    gas: Annotated[str, AfterValidator(Validators.valid_eth_expression("estimate"))]
    delay: Annotated[str | None, AfterValidator(Validators.valid_calc_decimal_value())] = None  # in seconds
    round_ndigits: int = 5
    proxies: Annotated[list[str], Field(default_factory=list), BeforeValidator(Validators.proxies())]
    wait_tx_timeout: int = 120
    log_debug: Annotated[Path | None, BeforeValidator(Validators.log_file())] = None
    log_info: Annotated[Path | None, BeforeValidator(Validators.log_file())] = None

    @property
    def from_addresses(self) -> list[str]:
        return [r.from_address for r in self.transfers]

    @model_validator(mode="after")
    def final_validator(self) -> Self:
        if not self.private_keys.contains_all_addresses(self.from_addresses):
            raise ValueError("private keys are not set for all addresses")

        for transfer in self.transfers:  # If value is not set for a transfer, then set it to the global value of the config.
            if not transfer.value and self.default_value:
                transfer.value = self.default_value
        for transfer in self.transfers:  # Check all transfers have a value.
            if not transfer.value:
                raise ValueError(f"{transfer.log_prefix}: value is not set")

        if self.token:
            if self.default_value:
                Validators.valid_token_expression("balance")(self.default_value)
            if self.value_min_limit:
                Validators.valid_token_expression()(self.value_min_limit)
        else:
            if self.default_value:
                Validators.valid_eth_expression("balance")(self.default_value)
            if self.value_min_limit:
                Validators.valid_eth_expression()(self.value_min_limit)

        if self.token:
            res = erc20.get_decimals(self.nodes, self.token, proxies=self.proxies, attempts=5)
            if isinstance(res, Err):
                fatal(f"can't get token decimals: {res.err}")
            self.token_decimals = res.ok

        return self


class TransferCmdParams(BaseConfigParams):
    print_balances: bool
    print_transfers: bool
    debug: bool
    skip_receipt: bool
    emulate: bool


def run(cmd_params: TransferCmdParams) -> None:
    config = Config.read_toml_config_or_exit(cmd_params.config_path)
    if cmd_params.print_config:
        cli_utils.print_config(config, exclude={"private_keys"}, count=None if cmd_params.debug else {"proxies"})
        sys.exit(0)

    rpc_helpers.check_nodes_for_chain_id(config.nodes, config.chain_id)

    if cmd_params.print_transfers:
        _print_transfers(config)
        sys.exit(0)

    if cmd_params.print_balances:
        _print_balances(config)
        sys.exit(0)

    _run_transfers(config, cmd_params)


def _run_transfers(config: Config, cmd_params: TransferCmdParams) -> None:
    mm_crypto_utils.init_logger(cmd_params.debug, config.log_debug, config.log_info)
    logger.info(f"transfer {cmd_params.config_path}: started at {utc_now()} UTC")
    logger.debug(f"config={config.model_dump(exclude={'private_keys'}) | {'version': cli_utils.get_version()}}")
    for i, transfer in enumerate(config.transfers):
        _transfer(transfer, config, cmd_params)
        if config.delay is not None and i < len(config.transfers) - 1:
            delay_value = mm_crypto_utils.calc_decimal_value(config.delay)
            logger.info(f"delay {delay_value} seconds")
            if not cmd_params.emulate:
                time.sleep(float(delay_value))
    logger.info(f"finished at {utc_now()} UTC")


def _transfer(t: Transfer, config: Config, cmd_params: TransferCmdParams) -> None:
    nonce = rpc_helpers.get_nonce(config.nodes, t.from_address, t.log_prefix)
    if nonce is None:
        return

    max_fee = rpc_helpers.calc_max_fee(config.nodes, config.max_fee, t.log_prefix)
    if max_fee is None:
        return

    if rpc_helpers.is_max_fee_limit_exceeded(max_fee, config.max_fee_limit, t.log_prefix):
        return

    gas = _calc_gas(t, config)
    if gas is None:
        return

    value = _calc_value(t, max_fee=max_fee, gas=gas, config=config)
    if value is None:
        return

    if not _check_value_min_limit(t, value, config):
        return

    priority_fee = calc_eth_expression(config.priority_fee)

    # emulate?
    if cmd_params.emulate:
        msg = f"{t.log_prefix}: emulate,"
        msg += f" value={_value_with_suffix(value, config)},"
        msg += f" max_fee={from_wei_str(max_fee, 'gwei', config.round_ndigits)},"
        msg += f" priority_fee={from_wei_str(priority_fee, 'gwei', config.round_ndigits)},"
        msg += f" gas={gas}"
        logger.info(msg)
        return

    tx_hash = _send_tx(transfer=t, nonce=nonce, max_fee=max_fee, priority_fee=priority_fee, gas=gas, value=value, config=config)
    if tx_hash is None:
        return

    status = "UNKNOWN"
    if not cmd_params.skip_receipt:
        logger.debug(f"{t.log_prefix}: waiting for receipt, tx_hash={tx_hash}")
        status = cli_utils.wait_tx_status(config.nodes, config.proxies, tx_hash, config.wait_tx_timeout)

    logger.info(f"{t.log_prefix}: tx_hash={tx_hash}, value={_value_with_suffix(value, config)},  status={status}")


def _calc_value(transfer: Transfer, max_fee: int, gas: int, config: Config) -> int | None:
    if config.token:
        return rpc_helpers.calc_erc20_value_for_address(
            nodes=config.nodes,
            value_expression=transfer.value,
            wallet_address=transfer.from_address,
            token_address=config.token,
            decimals=config.token_decimals,
            log_prefix=transfer.log_prefix,
        )
    return rpc_helpers.calc_eth_value_for_address(
        nodes=config.nodes,
        value_expression=transfer.value,
        address=transfer.from_address,
        gas=gas,
        max_fee=max_fee,
        log_prefix=transfer.log_prefix,
    )


def _check_value_min_limit(transfer: Transfer, value: int, config: Config) -> bool:
    """Returns False if the transfer should be skipped."""
    if config.value_min_limit:
        if config.token:
            value_min_limit = calcs.calc_token_expression(config.value_min_limit, config.token_decimals)
        else:
            value_min_limit = calcs.calc_eth_expression(config.value_min_limit)
        if value < value_min_limit:
            logger.info(f"{transfer.log_prefix}: value<value_min_limit, value={_value_with_suffix(value, config)}")
    return True


def _send_tx(
    *, transfer: Transfer, nonce: int, max_fee: int, priority_fee: int, gas: int, value: int, config: Config
) -> str | None:
    debug_tx_params = {
        "nonce": nonce,
        "max_fee": max_fee,
        "priority_fee": priority_fee,
        "gas": gas,
        "value": value,
        "to": transfer.to_address,
        "chain_id": config.chain_id,
    }
    logger.debug(f"{transfer.log_prefix}: tx_params={debug_tx_params}")

    if config.token:
        signed_tx = erc20.sign_transfer_tx(
            nonce=nonce,
            max_fee_per_gas=max_fee,
            max_priority_fee_per_gas=priority_fee,
            gas_limit=gas,
            private_key=config.private_keys[transfer.from_address],
            chain_id=config.chain_id,
            value=value,
            token_address=config.token,
            recipient_address=transfer.to_address,
        )
    else:
        signed_tx = sign_tx(
            nonce=nonce,
            max_fee_per_gas=max_fee,
            max_priority_fee_per_gas=priority_fee,
            gas=gas,
            private_key=config.private_keys[transfer.from_address],
            chain_id=config.chain_id,
            value=value,
            to=transfer.to_address,
        )
    res = rpc.eth_send_raw_transaction(config.nodes, signed_tx.raw_tx, attempts=5)
    if isinstance(res, Err):
        logger.info(f"{transfer.log_prefix}: tx error {res.err}")
        return None
    return res.ok


def _calc_gas(transfer: Transfer, config: Config) -> int | None:
    if config.token:
        return rpc_helpers.calc_gas(
            nodes=config.nodes,
            gas_expression=config.gas,
            from_address=transfer.from_address,
            to_address=config.token,
            data=erc20.encode_transfer_input_data(transfer.to_address, 1234),
            log_prefix=transfer.log_prefix,
        )
    return rpc_helpers.calc_gas(
        nodes=config.nodes,
        gas_expression=config.gas,
        from_address=transfer.from_address,
        to_address=transfer.to_address,
        value=123,
        log_prefix=transfer.log_prefix,
    )


def _print_transfers(config: Config) -> None:
    table = Table("n", "from_address", "to_address", "value", title="transfers")
    for count, transfer in enumerate(config.transfers, start=1):
        table.add_row(str(count), transfer.from_address, transfer.to_address, transfer.value)
    console = Console()
    console.print(table)


def _print_balances(config: Config) -> None:
    if config.token:
        headers = ["n", "from_address", "nonce", "eth", "t", "to_address", "nonce", "eth", "t"]
    else:
        headers = ["n", "from_address", "nonce", "eth", "to_address", "nonce", "eth"]
    table = Table(*headers, title="balances")
    with Live(table, refresh_per_second=0.5):
        for count, transfer in enumerate(config.transfers):
            from_nonce = _get_nonce_str(transfer.from_address, config)
            to_nonce = _get_nonce_str(transfer.to_address, config)

            from_eth_balance = _get_eth_balance_str(transfer.from_address, config)
            to_eth_balance = _get_eth_balance_str(transfer.to_address, config)

            from_token_balance = _get_token_balance_str(transfer.from_address, config) if config.token else ""
            to_token_balance = _get_token_balance_str(transfer.to_address, config) if config.token else ""

            if config.token:
                table.add_row(
                    str(count),
                    transfer.from_address,
                    from_nonce,
                    from_eth_balance,
                    from_token_balance,
                    transfer.to_address,
                    to_nonce,
                    to_eth_balance,
                    to_token_balance,
                )
            else:
                table.add_row(
                    str(count),
                    transfer.from_address,
                    from_nonce,
                    from_eth_balance,
                    transfer.to_address,
                    to_nonce,
                    to_eth_balance,
                )


def _get_nonce_str(address: str, config: Config) -> str:
    return str(rpc.eth_get_transaction_count(config.nodes, address, proxies=config.proxies, attempts=5).ok_or_err())


def _get_eth_balance_str(address: str, config: Config) -> str:
    return rpc.eth_get_balance(config.nodes, address, proxies=config.proxies, attempts=5).map_or_else(
        lambda err: err,
        lambda ok: from_wei_str(ok, "eth", config.round_ndigits),
    )


def _get_token_balance_str(address: str, config: Config) -> str:
    if not config.token:
        raise ValueError("token is not set")
    return erc20.get_balance(config.nodes, config.token, address, proxies=config.proxies, attempts=5).map_or_else(
        lambda err: err,
        lambda ok: from_wei_str(ok, "t", decimals=config.token_decimals, round_ndigits=config.round_ndigits),
    )


def _value_with_suffix(value: int, config: Config) -> str:
    if config.token:
        return from_wei_str(value, "t", config.round_ndigits, decimals=config.token_decimals)
    return from_wei_str(value, "eth", config.round_ndigits)



from mm_std import PrintFormat, print_json, print_plain

from mm_eth import erc20, rpc
from mm_eth.cli.cli_utils import public_rpc_url
from mm_eth.utils import from_wei_str


def run(rpc_url: str, wallet_address: str, token_address: str | None, wei: bool, print_format: PrintFormat) -> None:
    result: dict[str, object] = {}
    rpc_url = public_rpc_url(rpc_url)

    # nonce
    result["nonce"] = rpc.eth_get_transaction_count(rpc_url, wallet_address).ok_or_err()
    if print_format == PrintFormat.PLAIN:
        print_plain(f"nonce: {result['nonce']}")

    # eth balance
    result["eth_balance"] = (
        rpc.eth_get_balance(rpc_url, wallet_address).map(lambda x: str(x) if wei else from_wei_str(x, "eth")).ok_or_err()
    )
    if print_format == PrintFormat.PLAIN:
        print_plain(f"eth_balance: {result['eth_balance']}")

    if token_address:
        # token decimal
        result["token_decimal"] = erc20.get_decimals(rpc_url, token_address).ok_or_err()
        if print_format == PrintFormat.PLAIN:
            print_plain(f"token_decimal: {result['token_decimal']}")

        # token symbol
        result["token_symbol"] = erc20.get_symbol(rpc_url, token_address).ok_or_err()
        if print_format == PrintFormat.PLAIN:
            print_plain(f"token_symbol: {result['token_symbol']}")

        # token balance
        result["token_balance"] = (
            erc20.get_balance(rpc_url, token_address, wallet_address)
            .map(
                lambda x: str(x) if wei or not result["token_decimal"] else from_wei_str(x, "t", decimals=result["token_decimal"])  # type: ignore[arg-type]
            )
            .ok_or_err()
        )
        if print_format == PrintFormat.PLAIN:
            print_plain(f"token_balance: {result['token_balance']}")

    if print_format == PrintFormat.JSON:
        print_json(data=result)



import yaml
from mm_std import BaseConfig, fatal
from pydantic import StrictStr

from mm_eth import account, deploy
from mm_eth.cli import rpc_helpers
from mm_eth.cli.cli_utils import BaseConfigParams


class Config(BaseConfig):
    private_key: StrictStr
    nonce: int | None = None
    gas: StrictStr
    max_fee_per_gas: str
    max_priority_fee_per_gas: str
    value: str | None = None
    contract_bin: StrictStr
    constructor_types: StrictStr = "[]"
    constructor_values: StrictStr = "[]"
    chain_id: int
    node: str


class DeployCmdParams(BaseConfigParams):
    pass


def run(cli_params: DeployCmdParams) -> None:
    config = Config.read_toml_config_or_exit(cli_params.config_path)
    if cli_params.print_config:
        config.print_and_exit({"private_key"})

    constructor_types = yaml.full_load(config.constructor_types)
    constructor_values = yaml.full_load(config.constructor_values)

    sender_address = account.private_to_address(config.private_key)
    if sender_address is None:
        fatal("private address is wrong")

    nonce = rpc_helpers.get_nonce(config.node, sender_address)
    if nonce is None:
        fatal("can't get nonce")

    deploy.get_deploy_contract_data(config.contract_bin, constructor_types, constructor_values)



from eth_typing import HexStr
from rich.pretty import pprint
from web3 import Web3

from mm_eth.cli.cli_utils import public_rpc_url


def run(rpc_url: str, tx_hash: str, get_receipt: bool) -> None:
    rpc_url = public_rpc_url(rpc_url)
    w3 = Web3(Web3.HTTPProvider(rpc_url))
    transaction = w3.eth.get_transaction(HexStr(tx_hash))
    pprint(dict(transaction), expand_all=True)

    if get_receipt:
        receipt = w3.eth.get_transaction_receipt(HexStr(tx_hash))
        pprint(dict(receipt), expand_all=True)



import json

from mm_std import run_command


def test_mnemonic_cmd(mnemonic, address_0, private_0):
    cmd = f"mm-eth wallet mnemonic -m '{mnemonic}'"
    res = run_command(cmd)
    assert res.code == 0
    assert json.loads(res.stdout)["accounts"][0] == {"address": address_0, "private": private_0}



import json

from mm_std import run_command


def test_node_cmd(anvil):
    cmd = f"mm-eth node {anvil.rpc_url} -f json"
    res = run_command(cmd)
    assert res.code == 0
    assert json.loads(res.stdout)[0]["chain_id"] == 31337



from collections.abc import Callable

from mm_crypto_utils import AddressToPrivate, ConfigValidators, Transfer

from mm_eth.account import address_from_private, is_address
from mm_eth.constants import SUFFIX_DECIMALS

from . import calcs


class Validators(ConfigValidators):
    @staticmethod
    def valid_eth_expression(var_name: str | None = None) -> Callable[[str], str]:
        return ConfigValidators.valid_calc_int_expression(var_name, SUFFIX_DECIMALS)

    @staticmethod
    def valid_token_expression(var_name: str | None = None) -> Callable[[str], str]:
        return ConfigValidators.valid_calc_int_expression(var_name, {"t": 6})

    @staticmethod
    def valid_eth_or_token_expression(var_name: str | None = None) -> Callable[[str], str]:
        return ConfigValidators.valid_calc_int_expression(var_name, SUFFIX_DECIMALS | {"t": 6})

    @staticmethod
    def eth_transfers() -> Callable[[str], list[Transfer]]:
        return ConfigValidators.transfers(is_address, to_lower=True)

    @staticmethod
    def eth_private_keys() -> Callable[[str], AddressToPrivate]:
        return ConfigValidators.private_keys(address_from_private)

    @staticmethod
    def eth_address() -> Callable[[str], str]:
        return ConfigValidators.address(is_address, to_lower=True)

    @staticmethod
    def eth_addresses(unique: bool) -> Callable[[str], list[str]]:
        return ConfigValidators.addresses(unique, to_lower=True, is_address=is_address)


def is_valid_calc_function_args(value: str | None) -> bool:
    if value is None:
        return True
    try:
        calcs.calc_function_args(value)
        return True  # noqa: TRY300
    except ValueError:
        return False



import importlib.metadata
import time
from pathlib import Path
from typing import Literal

from mm_crypto_utils import Nodes, Proxies
from mm_std import BaseConfig, print_json
from pydantic import BaseModel

from mm_eth import rpc


def get_version() -> str:
    return importlib.metadata.version("mm-eth")


def public_rpc_url(url: str | None) -> str:
    if not url or url == "1":
        return "https://ethereum.publicnode.com"
    if url.startswith(("http://", "https://", "ws://", "wss://")):
        return url

    match url.lower():
        case "mainnet" | "1":
            return "https://ethereum.publicnode.com"
        case "sepolia" | "11155111":
            return "https://ethereum-sepolia-rpc.publicnode.com"
        case "opbnb" | "204":
            return "https://opbnb-mainnet-rpc.bnbchain.org"
        case "base" | "8453":
            return "https://mainnet.base.org"
        case "base-sepolia" | "84532":
            return "https://sepolia.base.org"
        case _:
            return url


class BaseConfigParams(BaseModel):
    config_path: Path
    print_config: bool


def print_config(config: BaseConfig, exclude: set[str] | None = None, count: set[str] | None = None) -> None:
    data = config.model_dump(exclude=exclude)
    if count:
        for k in count:
            data[k] = len(data[k])
    print_json(data)


def wait_tx_status(nodes: Nodes, proxies: Proxies, tx_hash: str, timeout: int) -> Literal["OK", "FAIL", "TIMEOUT"]:
    started_at = time.perf_counter()
    count = 0
    while True:
        res = rpc.get_tx_status(nodes, tx_hash, proxies=proxies, attempts=5)
        if res.is_ok():
            return "OK" if res.ok == 1 else "FAIL"

        time.sleep(1)
        count += 1
        if time.perf_counter() - started_at > timeout:
            return "TIMEOUT"



import json

from mm_eth.cli import calcs


def test_calc_function_args():
    res = calcs.calc_function_args('["xxx", random(100,200), 100, "aaa", random(1,3)]')
    assert json.loads(res)[1] >= 100



from mm_crypto_utils import Nodes, Proxies, random_node, random_proxy
from mm_std import Err, Ok, Result

from mm_eth.utils import get_w3


def get_name_with_retries(
    rpc_urls: Nodes, address: str, retries: int, timeout: float = 5, proxies: Proxies = None
) -> Result[str | None]:
    res: Result[str | None] = Err("not started yet")
    for _ in range(retries):
        res = get_name(random_node(rpc_urls), address, timeout=timeout, proxy=random_proxy(proxies))
        if res.is_ok():
            return res
    return res


def get_name(rpc_url: str, address: str, timeout: float = 5, proxy: str | None = None) -> Result[str | None]:
    try:
        w3 = get_w3(rpc_url, timeout=timeout, proxy=proxy)
        return Ok(w3.ens.name(w3.to_checksum_address(address)))  # type: ignore[union-attr]
    except Exception as e:
        return Err(e)



import json

from mm_std import run_command


def test_eth_balance_cmd(anvil, address_1):
    cmd = f"mm-eth balance {address_1} -u {anvil.rpc_url} -w -f json"
    res = run_command(cmd)
    assert res.code == 0
    assert json.loads(res.stdout)["eth_balance"] == "10000000000000000000000"
    assert json.loads(res.stdout)["nonce"] == 0



from mm_std import Ok

from mm_eth import ens


def test_get_name_exists(infura):
    address = "0xd8dA6BF26964aF9D7eEd9e03E53415D37aA96045"
    assert ens.get_name(infura(), address) == Ok("vitalik.eth")


def test_get_name_non_exists(infura):
    address = "0x743997F620846ab4CE946CBe3f5e5b5c51921D6E"  # random empty address
    assert ens.get_name(infura(), address) == Ok(None)



import random

import mm_crypto_utils
from mm_crypto_utils import VarInt

from mm_eth.constants import SUFFIX_DECIMALS


def calc_eth_expression(expression: str, var: VarInt | None = None) -> int:
    return mm_crypto_utils.calc_int_expression(expression, var=var, suffix_decimals=SUFFIX_DECIMALS)


def calc_token_expression(expression: str, token_decimals: int, var: VarInt | None = None) -> int:
    return mm_crypto_utils.calc_int_expression(expression, var=var, suffix_decimals={"t": token_decimals})


def calc_function_args(value: str) -> str:
    while True:
        if "random(" not in value:
            return value
        start_index = value.index("random(")
        stop_index = value.index(")", start_index)
        random_range = [int(v.strip()) for v in value[start_index + 7 : stop_index].split(",")]
        if len(random_range) != 2:
            raise ValueError("wrong random(from,to) template")
        rand_value = str(random.randint(random_range[0], random_range[1]))
        value = value[0:start_index] + rand_value + value[stop_index + 1 :]



from rich.live import Live
from rich.table import Table

from mm_eth import erc20, rpc
from mm_eth.utils import from_wei_str


def print_balances(
    rpc_nodes: list[str],
    addresses: list[str],
    *,
    token_address: str | None = None,
    token_decimals: int | None = None,
    round_ndigits: int = 5,
) -> None:
    table = Table(title="balances")
    table.add_column("n")
    table.add_column("address")
    table.add_column("nonce")
    table.add_column("balance, eth")
    if token_address is not None and token_decimals is not None:
        table.add_column("token, t")
    with Live(table, refresh_per_second=0.5):
        for count, address in enumerate(addresses):
            nonce = str(rpc.eth_get_transaction_count(rpc_nodes, address, attempts=5).ok_or_err())
            balance = rpc.eth_get_balance(rpc_nodes, address, attempts=5).map_or_else(
                lambda err: err,
                lambda ok: from_wei_str(ok, "eth", round_ndigits),
            )
            row: list[str] = [str(count), address, nonce, balance]
            if token_address is not None and token_decimals is not None:
                erc20_balance = erc20.get_balance(rpc_nodes, token_address, address, attempts=5).map_or_else(
                    lambda err: err,
                    lambda ok: from_wei_str(ok, "t", decimals=token_decimals, round_ndigits=round_ndigits),
                )
                row.append(erc20_balance)
            table.add_row(*row)



import json

from mm_std import run_command


def test_solc_cmd():
    cmd = "mm-eth solc tests/contracts/ERC20.sol -f json"
    res = run_command(cmd)
    assert res.code == 0
    assert json.loads(res.stdout)["bin"].startswith("60806040")



from mm_std import run_command


def test_private_key_cmd(private_0, address_0):
    cmd = f"mm-eth wallet private-key {private_0}"
    res = run_command(cmd)
    assert res.code == 0
    assert res.stdout.strip().lower() == address_0.lower()



from hexbytes import HexBytes
from mm_std import CustomJSONEncoder


class EthJsonEncoder(CustomJSONEncoder):
    def default(self, o: object) -> object:
        if isinstance(o, HexBytes):
            return o.hex()
        return super().default(o)


def json_default(o: object) -> str:
    if isinstance(o, HexBytes):
        return o.to_0x_hex()
    raise TypeError(f"Object of type {type(o).__name__} is not JSON serializable")
